'use strict';

const _ = require('lodash');
const GoogleCloudDatastore = require('@google-cloud/datastore');

const Base = require('./base');
const Model = require('../model');
const Types = require('paradox-utils').SchemaTypes;
const asyncp = require('paradox-asyncp');

const NotFoundError = require('../errors/not-found');
const UnknownError = require('../errors/unknown');
const PartitionError = require('../errors/partition');

class Datastore extends Base {

    /* -------------------------------------------------------------------------
     * Link
     * ---------------------------------------------------------------------- */

    getDataStore() {
        if(!this.datastore) {
            this.datastore = GoogleCloudDatastore(this.options.credentials);
        }

        return this.datastore;
    }

    isTransaction() {
        return this.getDataStore().constructor.name == 'Transaction';
    }

    getLink() {
        if(!this.link) {
            this.link = new Promise((resolve, reject) => {
                try {
                    resolve(this.getDataStore());

                } catch(err) {
                    reject(err);
                }
            });
        }

        return this.link;
    }

    define(name, fields, options) {
        let model = new Model(this.conn, name, fields, options);

        fields['key'] = {
            type: Types.KEY,
            primaryKey: true
        };

        return model;
    }

    /* -------------------------------------------------------------------------
     * Find
     * ---------------------------------------------------------------------- */

    findAll(options) {
        if(options.where && options.where.key) {
            return this
                .findByKey(options.where.key, options)
                .then(instance => {
                    return [instance];
                });
        }

        return this
            .createQueryFromOptions(options)
            .then((query) => {
                return query.run();
            })
            .then((data) => {
                return this.extractKeys(data[0]);
            });
    }

    findByKey(key, options) {
        return this
            .getLink()
            .then((link) => {
                key = this.key(key, options.table, options);
                return link
                    .get(key);
            })
            .then((data) => {
                if(!data[0]) {
                    return Promise.reject(new NotFoundError(new Error('empty response')));
                }

                return this.extractKey(data[0]);
            });
    }

    /* -------------------------------------------------------------------------
     * Count
     * ---------------------------------------------------------------------- */

    count(options) {
        return this.findAll(options)
            .then(rows => {
                return rows.length;
            });
    }

    /* -------------------------------------------------------------------------
     * Modify
     * ---------------------------------------------------------------------- */

    insert(data, options) {
        return this.update(data, options);
    }

    update(data, options) {
        if(options.where && options.where.key) {
            return this.updateByKey(options.where.key, data, options);

            // } else if(options.where && Object.keys(options.where).length > 0) {
            //     return this.updateByWhere(data, options);

        } else {
            return this.updateByKey(null, data, options);
        }
    }

    updateByWhere(data, options) {
        throw new Error('updated by where not implemented');
    }

    resolveObjPath(value, path, cb) {
        if(!path) {
            throw new Error('path is required');
        }

        if(typeof value === 'undefined') {
            return;
        }

        let cleanPath = path.replace(/\.\[/g, '[').replace(/\.+$/g, '').replace(/\.+/g, '.');

        if(_.isArray(value)) {
            _.forEach(value, subVal => {
                this.resolveObjPath(subVal, path + '[].', cb);
            });

        } else if(_.isPlainObject(value)) {
            _.forEach(value, (subVal, subKey) => {
                this.resolveObjPath(subVal, path + '.' + subKey + '.', cb);
            });

        } else {
            if(_.isInteger(value)) {
                cb && cb(cleanPath, true);

            } else if(_.isBoolean(value)) {
                cb && cb(cleanPath, true);

            } else if(_.isString(value)) {
                cb && cb(cleanPath, true);

                if(value.length >= 1500) {
                    cb && cb(cleanPath, false);
                }

            } else if(_.isDate(value)) {
                cb && cb(cleanPath, true);

            } else if(_.isBuffer(value)) {
                cb && cb(cleanPath, true);

            } else if(_.isNull(value)) {
                cb && cb(cleanPath, true);

            } else if(_.isNaN(value)) {
                cb && cb(cleanPath, true);
            }
        }


        return this;
    };

    updateByKey(key, data, options) {
        key = this.key(key, options.table, options);

        let noIndex = options.noIndex || options.model.options.noIndex || [];

        let fullIndex;

        if(typeof options.fullIndex != 'undefined') {
            fullIndex = options.fullIndex;

        } else if(typeof options.model.options.fullIndex != 'undefined') {
            fullIndex = options.model.options.fullIndex;

        } else {
            fullIndex = true;
        }

        /* -------------------------------------------------------------------------
         * Extract indexes
         * ---------------------------------------------------------------------- */

        let allIndexes = [];
        let noIndexable = [];

        _.forEach(data, (value, key) => {
            this.resolveObjPath(value, key, (path, indexable) => {
                if(path.match(/]$/)) {
                    return;
                }

                if(!indexable) {
                    noIndexable.push(path);

                } else {
                    allIndexes.push(path);
                }
            });
        });

        allIndexes = _.uniq(allIndexes);
        noIndexable = _.uniq(noIndexable);

        /* -------------------------------------------------------------------------
         * Model indexes
         * ---------------------------------------------------------------------- */

        let modelIndexes = [];

        _.forEach(options.model.options.indexes, index => {
            _.forEach(index.fields, field => {
                if(!field) {
                    // TODO: error?
                    return;
                }

                if(_.isPlainObject(field)) {
                    modelIndexes.push(field.attribute);

                } else {
                    modelIndexes.push(field);
                }
            });
        });

        /* -------------------------------------------------------------------------
         * No indexable
         * ---------------------------------------------------------------------- */

        let userNoIndexable = allIndexes
            .filter(index => {
                return !!noIndex
                    .find(noIndex => {
                        let initPart = index.substring(0, noIndex.length);
                        let next = index.substring(noIndex.length, noIndex.length + 1);

                        if(initPart == noIndex && (!next || next == '.' || next == '[')) {
                            return true;
                        }

                        return false;
                    });
            });

        let userIndexable = allIndexes
            .filter(index => {
                if(modelIndexes.length < 1) {
                    return true;
                }

                let foundIndexableValue = modelIndexes
                    .find(modelIndex => {
                        let initPart = index.substring(0, modelIndex.length);
                        let next = index.substring(modelIndex.length, modelIndex.length + 1);

                        if(initPart == modelIndex && (!next || next == '.' || next == '[')) {
                            return true;
                        }

                        return false;
                    });

                if(foundIndexableValue) {
                    return false;
                }

                return true;
            });

        /* -------------------------------------------------------------------------
         * Compile
         * ---------------------------------------------------------------------- */

        let propertiesToExclude;

        if(fullIndex) {
            // only no index the no indexable ones.
            propertiesToExclude = _.uniq(_.concat(userNoIndexable, noIndexable));

        } else {
            // exclude all properties except model index ones
            propertiesToExclude = _.uniq(_.concat(userIndexable, noIndexable));
        }

        let entity = {
            key: key,
            method: 'upsert',
            data: _.omit(data, ['key']),
            excludeFromIndexes: propertiesToExclude,
        };

        if(this.isTransaction()) {
            this.getDataStore().save(entity);

            data['key'] = this.plainKey(key);
            return Promise.resolve(data);
        }

        return this
            .getDataStore()
            .save(entity)
            .then(res => {
                data['key'] = this.plainKey(key);
                return data;
            });
    }

    runTransaction(cb) {
        let transaction = this.getDataStore().transaction();

        return transaction
            .run()
            .then(data => {
                let trans = data[0];

                let newAdapter = _.clone(this);
                newAdapter.datastore = trans;

                trans.key = this.getDataStore().key;

                return cb(newAdapter, trans)
                    .then(rows => {
                        return trans
                            .commit()
                            .then(() => rows);
                    })
                    .catch(err => {
                        return trans
                            .rollback()
                            .then(() => Promise.reject(err));
                    });
            });
    }

    findAndUpdate(data, options) {
        return this
            .runTransaction(adapter => {
                return adapter
                    .findAll(options)
                    .then(rows => {
                        // update them
                        return asyncp
                            .eachLimit(rows, 10, row => {
                                let key = row.key;
                                row = Object.assign(row, data);

                                return adapter.updateByKey(key, row, options);
                            });
                    });
            });
    }

    destroy(options) {
        if(options.where && options.where.key) {
            return this.destroyByKey(options.where.key, options);

        } else {
            return this
                .findAll(options)
                .then(rows => {
                    return asyncp
                        .eachLimit(rows, 10, row => {
                            return this.destroyByKey(row.key, options);
                        });
                });
        }
    }

    destroyByKey(key, options) {
        key = this.key(key, options.table, options);

        return this
            .getDataStore()
            .delete(key);
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    extractKey(object) {
        let ds = this.getDataStore();

        if(ds.KEY in object) {
            object['key'] = this.plainKey(object[ds.KEY]);
        }

        return object;
    }

    extractKeys(objects) {
        return objects.map((object) => {
            return this.extractKey(object);
        });
    }

    plainKey(key) {
        if(!key) {
            throw new UnknownError(new Error('key not fullfilled'));
        }

        let keyStr = null;

        if(typeof key != 'object') {
            keyStr = key;

        } else if(key.id) {
            keyStr = parseInt(key.id);

        } else if(key.name) {
            keyStr = key.name;

        } else if(key.path) {
            keyStr = key.path.pop();

        } else {
            throw new UnknownError(new Error('key is not valid'));
        }

        // Compatibility mode
        // keyStr.id = keyStr;
        // keyStr.name = keyStr;
        // keyStr.path = [keyStr];

        return keyStr;
    }

    getNs(options) {
        // Namespace
        let namespace = options.namespace || this.options.namespace;

        if(!namespace) {
            throw new UnknownError(new Error('a namespace is required'));
        }

        // Partition
        if(this.options.forcePartition && (!options.partition && options.partition !== false)) {
            throw new PartitionError(true);
        }

        if(options.partition) {
            namespace += '_' + options.partition;
        }

        return namespace;
    }

    key(key, table, options) {
        if(key && typeof key == 'object' && key.namespace && key.kind) {
            return key;
        }

        let namespace = this.getNs(options);

        // Key
        if(key) {
            if(typeof key == "string") { // TODO: No estoy muy seguro de esto
                if(key.match(/^\d{16}$/)) {
                    key = parseInt(key);
                }

            } else if(typeof key != "number") {
                key = key + "";
            }
        }

        return this.getDataStore().key({
            namespace: namespace,
            path: key ? [table, key] : [table],
        });
    }

    createQueryFromOptions(options) {
        return this
            .getLink()
            .then((link) => {
                let namespace = this.getNs(options);
                let table = options.table;
                let query = link.createQuery(namespace, table);

                /**
                 * @deprecated use attributes
                 */
                if(options.properties) {
                    options.attributes = options.properties;
                }

                if(options.attributes) {
                    this.applyAttributes(query, options.attributes);
                }

                if(options.where) {
                    this.applyWhere(query, options.where);
                }

                if(options.order) {
                    this.applyOrder(query, options.order);
                }

                if(options.limit) {
                    this.applyLimit(query, options.limit);
                }

                if(options.offset) {
                    this.applyOffset(query, options.offset);
                }

                return query;
            });
    }

    applyAttributes(query, attributes) {
        if(!_.isArray(attributes)) {
            attributes = [attributes];
        }

        query.select(attributes);

        return this;
    }

    applyWhere(query, where) {
        for(let key in where) {
            let value = where[key];

            if(typeof value == 'object') {
                for(let operator in value) {
                    let subvalue = value[operator];
                    let compatOperator = this.convertOperator(operator);

                    query.filter(key, compatOperator, subvalue);
                }

            } else {
                query.filter(key, value);
            }
        }

        return this;
    }

    applyOrder(query, order) {
        if(!_.isArray(order)) {
            order = [order];
        }

        if(order.length == 2 && ['DESC', 'ASC'].indexOf((order[1] + '').toUpperCase()) >= 0) {
            order = [order];
        }

        order.forEach((subOrder) => {
            if(_.isArray(subOrder) && subOrder.length == 1) {
                query.order(subOrder[0]);

            } else if(_.isArray(subOrder) && subOrder.length == 2) {
                query.order(subOrder[0], {
                    descending: (subOrder[1] + '').toUpperCase() == 'DESC',
                });

            } else if(typeof subOrder == 'string') {
                query.order(subOrder);

            } else {
                throw new Error('not available order method');
            }

        });

        return this;
    }

    applyLimit(query, limit) {
        query.limit(limit);

        return this;
    }

    applyOffset(query, offset) {
        query.offset(offset);

        return this;
    }

    convertOperator(operator) {
        /*
         $and: {a: 5}           // AND (a = 5)
         $or: [{a: 5}, {a: 6}]  // (a = 5 OR a = 6)
         $gt: 6,                // > 6
         $gte: 6,               // >= 6
         $lt: 10,               // < 10
         $lte: 10,              // <= 10
         $ne: 20,               // != 20
         $eq: 3,                // = 3
         $not: true,            // IS NOT TRUE
         $between: [6, 10],     // BETWEEN 6 AND 10
         $notBetween: [11, 15], // NOT BETWEEN 11 AND 15
         $in: [1, 2],           // IN [1, 2]
         $notIn: [1, 2],        // NOT IN [1, 2]
         $like: '%hat',         // LIKE '%hat'
         $notLike: '%hat'       // NOT LIKE '%hat'
         $iLike: '%hat'         // ILIKE '%hat' (case insensitive) (PG only)
         $notILike: '%hat'      // NOT ILIKE '%hat'  (PG only)
         $like: { $any: ['cat', 'hat']}
         // LIKE ANY ARRAY['cat', 'hat'] - also works for iLike and notLike
         $overlap: [1, 2]       // && [1, 2] (PG array overlap operator)
         $contains: [1, 2]      // @> [1, 2] (PG array contains operator)
         $contained: [1, 2]     // <@ [1, 2] (PG array contained by operator)
         $any: [2,3]            // ANY ARRAY[2, 3]::INTEGER (PG only)

         $col: 'user.organization_id' // = "user"."organization_id", with dialect specific column identifiers, PG in this example
         */

        // Operator (=, <, >, <=, >=). Default: =

        switch(operator) {
            case '$eq':
                return '=';
            case '$gt':
                return '>=';
            case '$gte':
                return '>=';
            case '$lt':
                return '<';
            case '$lte':
                return '<=';
        }

        throw new Error('operator "' + operator + '" not available in Google DataStore');
    }
}

module.exports = Datastore;