'use strict';

const _ = require('lodash');

const Nedb = require('nedb-promise');
const Base = require('./base');
const Model = require('../model');
const Types = require('paradox-utils').SchemaTypes;
const asyncp = require('paradox-asyncp');

const NotFoundError = require('../errors/not-found');
const UnknownError = require('../errors/unknown');
const PartitionError = require('../errors/partition');

class Datastore extends Base {

    /* -------------------------------------------------------------------------
     * Link
     * ---------------------------------------------------------------------- */

    getDataStore(table) {
        if(!table) {
            throw new Error('table not defined');
        }

        if(!this.datastore) {
            this.datastore = {};
        }

        if(!(table in this.datastore)) {
            this.datastore[table] = Nedb(this.options);
        }

        return this.datastore[table];
    }

    getLink(table) {
        if(!this.link) {
            this.link = new Promise((resolve, reject) => {
                try {
                    resolve(this.getDataStore(table));

                } catch(err) {
                    reject(err);
                }
            });
        }

        return this.link;
    }

    define(name, fields, options) {
        let model = new Model(this.conn, name, fields, options);

        fields['key'] = {
            type: Types.KEY,
            primaryKey: true
        };

        this.model = model;

        return model;
    }

    transformOptions(options) {
        options = _.clone(options);

        if(options && options.where && options.where.key) {
            options.where._id = options.where.key;
            delete options.where.key;
        }

        return options;
    }

    transformData(result, reverse) {
        if(_.isArray(result)) {
            return result.map(this.transformData.bind(this));
        }

        if(reverse) {
            if(result.key) {
                result._id = result.key;
                delete result.key;
            }

        } else {
            if(result._id) {
                result.key = result._id;
                delete result._id;
            }
        }


        return result;
    }

    /* -------------------------------------------------------------------------
     * Find
     * ---------------------------------------------------------------------- */

    findAll(options) {
        options = this.transformOptions(options);

        return this
            .getDataStore(options.table)
            .cfind(options.where)
            .sort(options.order)
            .skip(options.offset || 0)
            .limit(options.limit || 10)
            .exec()
            .then(instances => {
                return this.transformData(instances);
            });
    }

    /* -------------------------------------------------------------------------
     * Count
     * ---------------------------------------------------------------------- */

    count(options) {
        options = this.transformOptions(options);

        return this
            .getDataStore(options.table)
            .count(options.where);
    }

    /* -------------------------------------------------------------------------
     * Modify
     * ---------------------------------------------------------------------- */

    insert(data, options) {
        options = this.transformOptions(options);
        data = this.transformData(data, true);

        return this
            .getDataStore(options.table)
            .insert(data)
            .then(instances => {
                return this.transformData(instances);
            });
    }

    update(data, options) {
        options = this.transformOptions(options);
        data = this.transformData(data, true);

        if(options.where._id) {
            data._id = options.where._id;
        }

        return this
            .getDataStore(options.table)
            .update(options.where, data, {
                upsert: true,
                returnUpdatedDocs: true,
            })
            .then(changed => {
                return this.transformData(data);
            })
    }

    findAndUpdate(data, options) {
        options = this.transformOptions(options);
        data = this.transformData(data);

        throw new Error('not implemented');
    }

    destroy(options) {
        options = this.transformOptions(options);

        return this
            .getDataStore(options.table)
            .remove(options.where);
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    getNs(options) {
        // Namespace
        let namespace = options.namespace || this.options.namespace;

        if(!namespace) {
            throw new UnknownError(new Error('a namespace is required'));
        }

        // Partition
        if(this.options.forcePartition && (!options.partition && options.partition !== false)) {
            throw new PartitionError(true);
        }

        if(options.partition) {
            namespace += '_' + options.partition;
        }

        return namespace;
    }
}

module.exports = Datastore;