'use strict';

const _ = require('lodash');
const Types = require('paradox-utils').SchemaTypes;
const asyncp = require('paradox-asyncp');

const Instance = require('./instance');
const PartitionError = require('./errors/partition');

const Promise = require('bluebird');
const NotFoundError = require('./errors/not-found');
const __ = require('paradox-utils');

class Model {

    constructor(conn, name, fields, options) {
        options = _.defaults(options, {
            fullIndex: true,
            noIndex: [],
        });

        this.conn = conn;
        this.name = name;
        this.fields = {};
        this.attributes = this.fields;
        this.options = options;
        this.partitionName = null;

        this.setFields(fields);
        this.applyOptions(options);
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    /**
     * Returns a copy of the model with the partition embeded.
     *
     * @param name
     * @returns {*}
     */
    partition(name) {
        let newModel = this.clone();

        newModel.partitionName = name;
        newModel.options = _.cloneDeep(this.options);

        return newModel;
    }

    clone() {
        return _.clone(this);
    }

    /* -------------------------------------------------------------------------
     * Info
     * ---------------------------------------------------------------------- */

    applyOptions(options) {
        // Timestamps
        if(typeof options.timestamps === 'undefined' || options.timestamps) {
            if(!('created_at' in this.fields)) {
                this.fields['created_at'] = {
                    type: Types.DATETIME(),
                    internal: true,
                    validate: {
                        isDate: true,
                    },
                    get: function() {
                        return this.getDataValue('created_at') || new Date();
                    }
                };
            }

            if(!('updated_at' in this.fields) && this.options.adapter != 'bigquery') {
                this.fields['updated_at'] = {
                    type: Types.DATETIME(),
                    internal: true,
                    validate: {
                        isDate: true,
                    },
                    get: function() {
                        return new Date();
                    }
                };
            }
        }

        this.options = options;
        return this;
    }

    setFields(fields) {
        for(let key in fields) {
            let field = fields[key];

            if(typeof field == 'function') {
                fields[key] = {
                    type: field(),
                };
            }

            if(typeof field == 'object' && !('type' in field)) {
                let fType = fields[key];

                fields[key] = {
                    type: fType,
                };
            }

            if(typeof fields[key].type == 'function') {
                fields[key].type = fields[key].type();
            }
        }

        this.fields = fields;
        this.attributes = fields;
        return this;
    }

    getFields() {
        return this.fields;
    }

    getFieldNames() {
        return this.fields.keys();
    }

    /* -------------------------------------------------------------------------
     * Sync
     * ---------------------------------------------------------------------- */

    async exists(options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        return this.conn.getAdapter().exists(options);
    }

    async sync(options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        return this.conn.getAdapter().sync(options);
    }

    /* -------------------------------------------------------------------------
     * Adapter
     * ---------------------------------------------------------------------- */

    getAdapter() {
        return this.conn.getAdapter();
    }

    /* -------------------------------------------------------------------------
     * Build
     * ---------------------------------------------------------------------- */

    /**
     * @param data
     * @param options
     * @return {Instance}
     */
    build(data, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        data = __.object.simplify(data);

        let record = new Instance(this.conn, this, data, options);
        return record;
    }

    /**
     * @param data
     * @param options
     * @returns {Promise}
     */
    create(data, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        return this.build(data, options).save(options);
    }

    bulkCreate(records, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        records = records.map(r => this.build(r, options));

        return this.saveInstances(records, options);
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    checkOptions(options) {
        options = _.defaults(options, {
            partition: null,
        });

        if((!options.partition && options.partition !== false) && (this.partitionName || this.partitionName === false)) {
            options.partition = this.partitionName;
        }

        if(this.options.forcePartition && (!options || (!options.partition && options.partition !== false))) {
            return Promise.reject(new PartitionError());
        }
    }

    /* -------------------------------------------------------------------------
     * Find
     * ---------------------------------------------------------------------- */

    findOrCreate(options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        return this
            .findOne(options)
            .catch(err => {
                if(err.code == 404) {
                    return null;
                }

                return Promise.reject(err);
            })
            .then(instance => {
                if(!instance) {
                    let data = _.defaults(
                        {},
                        options.defaults,
                        options.where // TODO: solo relaciones directas, sin $ne, $eq, etc
                    );

                    return this
                        .create(data, options)
                        .then(instance => {
                            return [instance, true];
                        });
                }

                return [instance, false];
            });
    }

    /**
     * @param options
     * @returns {*}
     */
    findOrBuild(options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        return this
            .findOne(options)
            .catch(err => {
                if(err.code == 404) {
                    return null;
                }

                return Promise.reject(err);
            })
            .then(instance => {
                if(!instance) {
                    let data = _.defaults(
                        {},
                        options.defaults,
                        options.where // TODO: solo relaciones directas, sin $ne, $eq, etc
                    );

                    let instance = this.build(data, options);
                    return [instance, true];
                }

                return [instance, false];
            });
    }

    /**
     * @alias findOrBuild
     *
     * @param options
     * @returns {*}
     */
    findOrInitialize(options) {
        return this.findOrBuild(options);
    }

    /**
     * @deprecated use findOrBuild
     *
     * @param options
     * @returns {*}
     */
    findOrInstanciate(options) {
        return this.findOrBuild(options);
    }

    metadata(options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        return this.conn.getAdapter().metadata(options);
    }

    explain(query, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        return this.conn.getAdapter().explain(query, options);
    }

    query(query, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let result = this.conn.getAdapter().query(query, options);

        if(options.iterate) {
            result.iterateCallback = row => {
                return this.build(row, options);
            };

            return result;

        } else {
            return result
                .then((rows) => {
                    return rows.map((row) => {
                        return this.build(row, options);
                    });
                });
        }
    }

    findAndUpdate(data, options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        data = __.object.simplify(data);

        let result = this.conn.getAdapter().findAndUpdate(data, options);

        if(options.iterate) {
            result.iterateCallback = row => {
                return this.build(row, options);
            };

            return result;

        } else {
            return result
                .then((rows) => {
                    return rows.map((row) => {
                        return this.build(row, options);
                    });
                });
        }
    }

    findOneAndUpdate(data, options) {
        options = options || {};
        options.limit = 1;

        return this
            .findAndUpdate(data, options)
            .then((rows) => {
                if(rows.length > 0) {
                    return rows[0];
                }

                return Promise.reject(new NotFoundError());
            });
    }

    findAll(options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        let result = this.conn.getAdapter().findAll(options);

        result = Promise.resolve(result);

        if(options.iterate) {
            result.iterateCallback = row => {
                return this.build(row, options);
            };

            return result;

        } else {
            return result
                .catch(err => {
                    if(err.code == 404) {
                        return [];
                    }

                    return Promise.reject(err);
                })
                .then((rows) => {
                    return rows.map((row) => {
                        return this.build(row, options);
                    });
                });
        }
    }

    findOne(options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        options['limit'] = 1;

        return this
            .findAll(options)
            .then((instances) => {
                if(instances && instances.length > 0) {
                    return instances[0];
                }

                if(this.options.emptyThrowException) {
                    return Promise.reject(new NotFoundError());
                }

                return null;
            })
            .catch(err => {
                console.error(err);
                return Promise.reject(err);
            })
    }

    /**
     * @alias findById
     *
     * @param id
     * @param options
     * @returns {*}
     */
    findByPrimary(id, options) {
        return this.findById(id, options);
    }

    findById(id, options) {
        options = options || {};

        let check = this.checkOptions(options);
        if(check) return check;

        let idField = this.getPrimaryKey();

        if(!idField) {
            throw new Error('primary key field not defined');
        }

        let where = {};
        where[idField] = id;

        return this
            .findOne(Object.assign({
                where
            }, options));
    }

    count(options) {
        options = _.defaultsDeep(options || {}, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        return this.conn.getAdapter()
            .count(options);
    }

    /* -------------------------------------------------------------------------
     * Special
     * ---------------------------------------------------------------------- */

    /**
     * @deprecated use findById
     *
     * @param key
     * @param options
     * @returns {*}
     */
    findByKey(key, options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        return this.conn.getAdapter()
            .findById(key, options)
            .then((row) => {
                if(!row) {
                    return null;
                }

                return this.build(row, options);
            });
    }

    /* -------------------------------------------------------------------------
     * Copy
     * ---------------------------------------------------------------------- */

    copy(remoteModel, options) {
        options = _.defaults(options, {
            concurrency: 1,
        });

        let originaOptions = _.clone(options);

        return this
            .findAll(options)
            .then(entities => {
                return asyncp
                    .eachLimit(entities, options.concurrency, entity => {
                        return remoteModel
                            .create(entity, _.clone(originaOptions));
                    })
                    .then(() => {
                        return {
                            total: entities.length,
                        };
                    });
            });
    }

    /* -------------------------------------------------------------------------
     * Modify
     * ---------------------------------------------------------------------- */

    destroy(options) {
        options = _.defaultsDeep(options, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        return this.conn.getAdapter()
            .destroy(options);
    }

    destroyInstance(instance, options) {
        options = _.defaultsDeep(options || {}, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        let idField = this.getPrimaryKey();

        if(!idField) {
            throw new Error('primary key field not defined and necessary');
        }

        if(idField in instance && instance[idField]) {
            let where = {};
            where[idField] = instance[idField];

            options = _.defaultsDeep(options, {
                where
            });

            return this.conn.getAdapter()
                .destroy(options)
                .then(() => {
                    return instance;
                });

        } else {
            return Promise.resolve();
        }
    }

    reloadInstance(instance, options) {
        options = _.defaultsDeep(options || {}, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        let idField = this.getPrimaryKey();

        if(!idField) {
            throw new Error('primary key field not defined and necessary');
        }

        if(idField in instance && instance[idField]) {
            let where = {};
            where[idField] = instance[idField];

            options = _.defaultsDeep(options, {
                where
            });

            return this
                .findOne(options)
                .then(node => {
                    if(!node) {
                        return instance;
                    }

                    let newValues = node.getValues();
                    instance.applyValues(newValues, true);
                    return instance;
                });

        } else {
            // not necessary to reload, not in db.
            // TODO: reject or resolve?
            return Promise.resolve(instance);
        }
    }

    saveInstances(instances, options) {
        options = _.defaultsDeep(options || {}, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        let records = instances.map(instance => {
            let data = instance.getValues();

            // Aplicar modificadores
            this.applyDataMods(data, options);

            // Validar
            this.validateData(data);

            return data;
        });

        return this.conn.getAdapter()
            .bulkInsert(records, options)
            .then(rows => {
                return rows.map((row) => {
                    return this.build(row, options);
                });
            });
    }

    saveInstance(instance, options) {
        options = _.defaultsDeep(options || {}, {
            table: this.name,
            model: this,
        });

        let check = this.checkOptions(options);
        if(check) return check;

        let data = instance.getValues();

        // Aplicar modificadores
        this.applyDataMods(data, options);

        // Validar
        this.validateData(data);

        let idField = this.getPrimaryKey();

        if(!idField) {
            throw new Error('primary key field not defined and necessary');
        }

        if(idField in data && data[idField]) {
            let where = {};
            where[idField] = data[idField];
            delete data[idField];

            options = _.defaultsDeep(options, {
                where
            });

            return this.conn.getAdapter()
                .update(data, options)
                .then((data) => {
                    return this.build(data, options);
                });

        } else {
            return this.conn.getAdapter()
                .insert(data, options)
                .then((data) => {
                    return this.build(data, options);
                });
        }
    }

    applyDataMods(data, options) {
        let fields = this.getFields();

        for(let fieldName in fields) {
            let field = fields[fieldName];

            // Default value
            if(!(fieldName in data)) {
                if('defaultValue' in field) {
                    data[fieldName] = field['defaultValue'];

                } else {
                    data[fieldName] = null;
                }
            }

            // Getter / Setter
            data[fieldName] = this.applyPropertyGetter(fieldName, data, data[fieldName], options);
        }

        return data;
    }

    applyPropertyGetter(name, data, value, options) {
        let field = this.getFieldByName(name);

        if(!field) {
            throw new Error('field with name "' + name + '" not found in model.');
        }

        if('get' in field) {
            value = field['get'].apply({
                getDataValue: (valueKey) => {
                    return valueKey in data ? data[valueKey] : undefined;
                },
                setDataValue: (valueKey, valueVal) => {
                    data[valueKey] = valueVal;
                    return valueVal;
                }

            }, [value]);
        }

        return value;
    }

    getFieldByName(name) {
        if(!(name in this.fields)) {
            throw new Error('field "' + name + '" not found');
        }

        return this.fields[name];
    }

    validateData(data) {

        return this;
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    getPartitionKey() {
        for(let key in this.attributes) {
            if('partitionKey' in this.attributes[key] && this.attributes[key].partitionKey) {
                return key;
            }
        }

        return null;
    }

    getPrimaryKey() {
        let idField = null;

        for(let key in this.attributes) {
            if(key == 'id') { // TODO: VALORAR ESTO
                idField = key;
            }

            if('primaryKey' in this.attributes[key] && this.attributes[key].primaryKey) {
                return key;
            }
        }

        return idField;
    }

    isPromise(value) {
        return value && (value instanceof Promise || (typeof value === 'object' && typeof value.then === 'function'));
    }
}

module.exports = Model;