'use strict';

const chai = require('chai');
const mocha = require('mocha');
const should = require('should');

const Dblize = require('../index');

describe.skip('ElasticSearch', function() {
    this.timeout(10 * 1000);

    let conn = null;
    let TestModel = null;
    let TestModelPartition = null;

    let partition = null;

    /* -------------------------------------------------------------------------
     * Before
     * ---------------------------------------------------------------------- */

    before(function() {
        conn = new Dblize({
            adapter: 'elasticsearch',
            credentials: {
                host: process.env.CRED_ELASTICSEARCH,
            },
            namespace: 'test',
            forcePartition: true,
        });
    });

    /* -------------------------------------------------------------------------
     * Create
     * ---------------------------------------------------------------------- */

    const validateRecord = instance => {
        (instance).should.have.property('id').which.is.an.String();
        (instance).should.have.property('title').which.is.an.String();
        (instance).should.have.property('episodes').which.is.an.Number();
        (instance).should.have.property('success').which.is.a.Boolean();
        (instance).should.have.property('created_at').which.is.a.Date();
        (instance).should.have.property('updated_at').which.is.a.Date();

        (instance.id).should.equal('1');
        (instance.title).should.equal('this is a test');
        (instance.episodes).should.equal(4);
        (instance.success).should.be.True();
    };

    describe('Models', () => {
        it('define a model', () => {
            partition = 'demo';

            TestModel = conn
                .define('resource.image', {
                    title: {
                        type: Dblize.STRING,
                    },
                    episodes: {
                        type: Dblize.NUMBER,
                    },
                    success: {
                        type: Dblize.BOOLEAN,
                    }
                });

            TestModelPartition = TestModel.partition('demo');
        });

        it('should require partition', () => {
            return TestModel
                .count()
                .should.be.rejectedWith({
                    message: 'partition not defined',
                });
        });

        it('should not require partition', () => {
            return TestModel
                .count({
                    partition: false,
                })
                .should.be.fulfilledWith(0);
        });

        it('should not require partition', () => {
            return TestModelPartition
                .count()
                .should.not.be.rejectedWith({
                    message: 'partition not defined',
                });
        });

        it('should not require partition', () => {
            return TestModelPartition
                .count()
                .should.be.fulfilledWith(0);
        });

        it('should sync the model', () => {
            return TestModel.sync({
                partition: partition
            });
        });

        it('should search one and not find nothing', () => {
            return TestModel
                .findById(1, {
                    partition: partition,
                })
                .should.be.rejectedWith({
                    code: 404,
                })
        });

        it('should insert a record with id 1', () => {
            return TestModel
                .create({
                    id: 1,
                    title: 'this is a test',
                    episodes: 4,
                    success: true,
                }, {
                    partition: partition
                })
                .then(instance => {
                    return validateRecord(instance);
                });
        });

        it('should search one', () => {
            return TestModel
                .findById(1, {
                    partition: partition
                })
                .then(instance => {
                    return validateRecord(instance);
                })
        });

        it('should count all entities', () => {
            return TestModel
                .count({
                    partition: partition
                })
                .should.be.fulfilledWith(1);
        });

        it('should query and count all', () => {
            return TestModel
                .count({
                    where: {
                        q: 'title:*'
                    },
                    partition: partition,
                })
                .should.be.fulfilledWith(1);
        });

        it('should search all', () => {
            return TestModel
                .findAll({
                    partition: partition
                })
                .then(instances => {
                    let instance = instances[0];
                    return validateRecord(instance);
                });
        });

        it('should query all', () => {
            return TestModel
                .query({
                    q: 'title:test',
                }, {
                    partition: partition,
                })
                .then(instances => {
                    let instance = instances[0];
                    return validateRecord(instance);
                });
        });

        it('should query by title', () => {
            return TestModel
                .findAll({
                    where: {
                        query: {
                            match: {
                                title: 'test'
                            }
                        }
                    },
                    partition: partition,
                })
                .then(instances => {
                    let instance = instances[0];
                    return validateRecord(instance);
                });
        });

        it('should delete all', () => {
            return TestModel
                .destroy({
                    where: {
                        id: 1,
                    },
                    partition: partition,
                })
                .then(res => {
                    (res.found).should.be.True();
                    return TestModel.count({
                        partition: partition,
                    });
                })
                .should.be.fulfilledWith(0);
        });

        it('should delete the index', () => {
            return TestModel
                .getAdapter()
                .getLink()
                .indices.delete({
                    index: 'test_demo_resource',
                });
        });
    });
});