'use strict';

const _ = require('lodash');
const EventEmitter = require('events').EventEmitter;

class Restful extends EventEmitter {

    constructor(name, model) {
        super();

        this.name = name;
        this.db = null;
        this.model = model;

        this.filters = {};
        this.order = {};
    }

    registerRouter(router) {
        this.router = router;

        router.route('/' + this.name)
            .get(this.routerList.bind(this))
            .post(this.routerCreate.bind(this));

        router.route('/' + this.name + '/:id')
            .get(this.routerShow.bind(this))
            .put(this.routerUpdate.bind(this))
            .patch(this.routerUpdate.bind(this))
            .delete(this.routerDestroy.bind(this));

        return this;
    }

    isPromise(value) {
        return value && (this.isPromise(value) || (typeof value === 'object' && typeof value.then === 'function'));
    }

    registerMethod(name, cb) {
        this.router.route('/' + this.name + '/' + name)
            .all((req, res) => {
                try {
                    let result = cb.apply(this);

                    if(this.isPromise(result)) {
                        result.then((promiseResult) => {
                            this.outputSuccess(promiseResult, res);
                            this.emit('method', name);
                        });

                    } else {
                        this.outputSuccess(result, res);
                        this.emit('method', name);
                    }

                } catch(err) {
                    this.outputError(err, res);
                }
            });

        return this;
    }

    registerDetailMethod(name, cb) {
        this.router.route('/' + this.name + '/' + name + '/:id')
            .get((req, res) => {
                try {
                    let id = this.extractRequestId(req);
                    let result = cb.apply(this, [id]);

                    if(this.isPromise(result)) {
                        result.then((promiseResult) => {
                            this.outputSuccess(promiseResult, res);
                            this.emit('method', name);
                        })

                    } else {
                        this.outputSuccess(result, res);
                        this.emit('method', name);
                    }

                } catch(err) {
                    this.outputError(err, res);
                }
            });

        return this;
    }

    addFilter(key, value) {
        this.filters[key] = value;
        return this;
    }

    addOrder(key, ascending) {
        this.order[key] = ascending;
        return this;
    }

    /*
     * Router methods
     */

    routerList(req, res) {
        let filters = this.extractRequestFilters(req);
        let limit = this.extractRequestLimit(req);
        let fields = this.extractRequestFields(req);
        let offset = this.extractRequestOffset(req);
        let order = this.extractRequestOrder(req);

        this.fetchCountRows(filters)
            .then((total) => {
                res.setHeader('X-Total-Count', total);
                res.setHeader('Access-Control-Expose-Headers', 'X-Total-Count');

                return this.fetchRows(filters, limit, offset, order, fields)
                    .then((rows) => {
                        rows.forEach((row) => {
                            this.emit('list.item', row);
                        });

                        this.outputSuccess(rows, res);
                        this.emit('list');
                    });
            })
            .catch((err) => {
                this.outputError(err, res);
            });
    }

    routerCreate(req, res) {
        let params = this.extractRequestParams(req);

        this.insertRow(params)
            .then((row) => {
                this.outputSuccess(row, res);
                this.emit('insert');
                this.emit('modify');
            })
            .catch((err) => {
                this.outputError(err, res);
            });
    }

    routerShow(req, res) {
        let id = this.extractRequestId(req);

        this.fetchRowById(id)
            .then((row) => {
                this.outputSuccess(row, res);
                this.emit('fetch');
            })
            .catch((err) => {
                this.outputError(err, res);
            });
    }

    routerUpdate(req, res) {
        let id = this.extractRequestId(req);
        let params = this.extractRequestParams(req);

        let method = req.method === 'PATCH' ? this.updateRowById.bind(this) : this.replaceRowById.bind(this);

        method(id, params)
            .then((row) => {
                this.outputSuccess(row, res);
                this.emit('update');
                this.emit('modify');
            })
            .catch((err) => {
                this.outputError(err, res);
            });
    }

    routerDestroy(req, res) {
        let id = this.extractRequestId(req);

        this.deleteRowById(id)
            .then((row) => {
                this.outputSuccess(row, res);
                this.emit('delete');
                this.emit('modify');
            })
            .catch((err) => {
                this.outputError(err, res);
            });
    }

    /*
     * Parse input
     */

    extractRequestId(req) {
        return this.convertValueToNative(req.params.id);
    }

    extractRequestOrder(req) {
        let sortKey = req.query.order || 'created_at';
        let sortDir = req.query.order_dir || 'asc';

        if(sortDir === 'asc') {
            sortDir = 1;
        } else {
            sortDir = -1;
        }

        let sort = {};
        sort[sortKey] = sortDir;

        return sort;
    }

    extractRequestFilters(req) {
        let filters = {};

        if(req.query['_filters']) {
            let _filters = JSON.parse(req.query['_filters']);

            for(let key in _filters) {
                filters[key] = this.convertValueToNative(_filters[key]);
            }
        }
        // _.defaults(filters, this.filters);

        return filters;
    }

    extractRequestFields(req) {
        let searchFor = ['fields'];
        let fields = null;

        for(let i = 0; i < searchFor.length; i++) {
            let key = searchFor[i];
            let dashkey = '_' + searchFor[i];

            if(key in req.query) {
                fields = req.query[key];

            } else if(dashkey in req.query) {
                fields = req.query[dashkey];
            }
        }

        return fields ? fields.split(/\s*,\s*/) : null;
    }

    extractRequestLimit(req) {
        let searchFor = ['limit', 'end'];
        let limit = 1000;

        for(let i = 0; i < searchFor.length; i++) {
            let key = searchFor[i];
            let dashkey = '_' + searchFor[i];

            if(key in req.query) {
                limit = parseInt(req.query[key]);

            } else if(dashkey in req.query) {
                limit = parseInt(req.query[dashkey]);
            }
        }

        return limit <= 0 ? 1000 : limit;
    }

    extractRequestOffset(req) {
        let searchFor = ['start', 'offset', 'from'];
        let offset = 0;

        for(let i = 0; i < searchFor.length; i++) {
            let key = searchFor[i];
            let dashkey = '_' + searchFor[i];

            if(key in req.query) {
                offset = parseInt(req.query[key]);

            } else if(dashkey in req.query) {
                offset = parseInt(req.query[dashkey]);
            }
        }

        return offset;
    }

    extractRequestParams(req) {
        let params = {};

        for(let key in req.body) {
            if(!req.body.hasOwnProperty(key)) {
                continue;
            }

            params[key] = this.convertValueToNative(req.body[key])
        }

        return params;
    }

    /*
     * Db methods
     */

    fetchCountRows(filters) {
        return new Promise((resolve, reject) => {
            resolve(1000);
        });
    }

    fetchRows(filters, limit, offset, order, fields) {
        return new Promise((resolve, reject) => {
            let query = this.db.query();

            if(Object.keys(this.order).length > 0) {
                for(let column in this.order) {
                    query.order(column, {
                        descending: !this.order[column]
                    });
                }
            }

            if(Object.keys(this.filters).length > 0) {
                for(let key in this.filters) {
                    query.filter(key, '=', this.filters[key]); // support: =, <, >, <=, >=
                }
            }

            // if(fields && fields.length > 0) {
            //     fields = _.deleteArrayItem(fields, '_key');
            //     query.select(fields);
            // }

            query
                .limit(limit)
                .offset(offset)
                .run((err, docs) => {
                    if(err) reject(err);
                    else {
                        docs.forEach((doc) => {
                            this.convertDbId(doc);
                        });

                        docs = this.applyFilters(docs, filters);
                        docs = this.applySort(docs, order);

                        resolve(docs);
                    }
                });
        });
    }

    applySort(rows, order) {
        rows.sort((rowA, rowB) => {
            for(let column in order) {
                let dir = order[column];

                let valA = _.queryArray(rowA, column) + "";
                let valB = _.queryArray(rowB, column) + "";

                if(valA > valB) return dir < 0 ? -1 : +1;
                if(valA < valB) return dir < 0 ? +1 : -1;
            }

            return 0;
        });

        return rows;
    }

    applyFilters(rows, filters) {
        let result = [];

        rows.forEach((row) => {
            let found = true;

            for(let key in filters) {
                let filterBy = filters[key] + "";

                filterBy = filterBy.replace(/ 00:00:00/, '');
                filterBy = filterBy.trim();

                let regex = new RegExp(filterBy, 'i');
                let value = _.queryArray(row, key) + "";

                if(!value.match(regex)) {
                    found = false;
                }
            }

            if(found) {
                result.push(row);
            }
        });


        return result;
    }

    fetchRowById(id) {
        return new Promise((resolve, reject) => {
            let ds = this.db.getDataStore();
            let key = this.db.key(id);

            ds.get(key, (err, entity) => {
                if(err) reject(err);
                else {
                    if(!entity) {
                        reject(new Error('id not found'));

                    } else {
                        resolve(this.convertDbId(entity));
                    }
                }
            });
        });
    }

    insertRow(data) {
        return new Promise((resolve, reject) => {
            data['created_at'] = new Date();
            data['updated_at'] = new Date();

            delete data['_key'];

            let key = this.db.key(data._key);
            let entity = {
                key: key,
                method: 'insert',
                data: data,
            };

            this.db.getDataStore().save(entity, (err) => {
                if(err) reject(err);
                else {
                    resolve(this.convertDbId(entity.data, key));
                }
            });
        });
    }

    updateRowById(id, data) {
        return new Promise((resolve, reject) => {
            data['updated_at'] = new Date();

            delete data['_key'];

            let key = this.db.key(id);
            let entity = {
                key: key,
                method: 'update',
                data: data,
            };

            this.db.getDataStore().save(entity, (err) => {
                if(err) reject(err);
                else {
                    resolve(this.convertDbId(entity.data, key));
                }
            });
        });
    }

    replaceRowById(id, data) {
        return new Promise((resolve, reject) => {
            data['updated_at'] = new Date();

            delete data['_key'];

            let key = this.db.key(id);
            let entity = {
                key: key,
                method: 'upsert',
                data: data,
            };

            this.db.getDataStore().save(entity, (err) => {
                if(err) reject(err);
                else {
                    resolve(this.convertDbId(entity.data, key));
                }
            });
        });
    }

    deleteRowById(id) {
        return new Promise((resolve, reject) => {
            let ds = this.db.getDataStore();
            let key = this.db.key(id);

            ds.delete(key, (err, entity) => {
                if(err) reject(err);
                else {
                    if(!entity) {
                        reject(new Error('id not found'));

                    } else {
                        resolve();
                    }
                }
            });
        });
    }

    /*
     * Utils
     */

    convertDbId(row, key) {
        if(!row[this.db.getDataStore().KEY] && !key) {
            return row;
        }

        row._key = key ? key.id : (row[this.db.getDataStore().KEY].id || row[this.db.getDataStore().KEY].name);

        return row;
    }

    convertValueToNative(value) {
        if(typeof value === 'string') {
            if(value === '' || value.trim() !== value || (value.length > 1 && value[0] === '0')) {
                return value
            } else if(value === 'true' || value === 'false') {
                return value === 'true'
            } else if(!isNaN(+value)) {
                return +value
            }
        }

        return value
    }

    /*
     * Result
     */

    outputSuccess(result, res) {
        res.status(200).json(result);
        return this;
    }

    outputError(err, res) {
        res.status(500).json(this.formatError(err));
        return this;
    }

    formatError(err) {
        return {
            error: err.message,
            name: err.name,
            filename: err.fileName,
            line: err.lineNumber,
            column: err.columnNumber,
            stack: err.stack ? err.stack.split(/\n/) : []
        };
    }

}

module.exports = Restful;